use axum::{routing::get, Router};
use tokio::sync::Mutex;

use std::sync::Arc;

use crate::{state::AppState, storages::StoragePool};

mod api;
mod config;
mod crypt;
mod routes;
mod state;
mod storages;

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    tracing_subscriber::fmt::init();

    let config = config::load().await;
    let secret_key = config.secret_key.clone();

    let mut storage_pool = StoragePool::from_config(config.storages);
    storage_pool.init().await?;

    let app_state = AppState::new(storage_pool, secret_key);
    let routes = match app_state.secret_key {
        Some(_) => get(routes::handle_secure),
        None => get(routes::handle_unsecure),
    };

    let app_state_th = Arc::new(Mutex::new(app_state));

    tokio::spawn(launch_storages_tick(app_state_th.clone()));

    let app = Router::new()
        .route("/*src", routes)
        .with_state(app_state_th);

    let listener = tokio::net::TcpListener::bind("0.0.0.0:3000").await?;
    axum::serve(listener, app).await?;

    Ok(())
}

async fn launch_storages_tick(app_state: Arc<Mutex<AppState>>) {
    loop {
        tokio::time::sleep(std::time::Duration::from_secs(1)).await;

        let mut app_state_locked = app_state.lock().await;
        let results = app_state_locked.storage_pool.tick().await;
        let errs = results
            .into_iter()
            .filter(anyhow::Result::is_err)
            .map(anyhow::Result::unwrap_err);

        for err in errs {
            println!("error while tick {}", err);
        }
    }
}
